#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <vector>
#include <set>
#include <map>

using namespace std;

#include "bt_peer.h"
#include "bt_tracker.h"
#include "bt_subtracker.h"
#include "event.h"
#include "memory_pool.h"
#include "topology.h"
#include "util.h"

/******************************************************************************
 *                                   NOTES                                    *
 ******************************************************************************
 * The bt_peer id corresponds to topology nodes in the following way.         *
 * University nodes have ids 0 to _num_univ_nodes-1.                          *
 * Cable modem leaf nodes have ids from _num_univ_nodes to                    *
 * _num_cable_leaf_nodes-1.                                                   *
 * University nodes are arranged in an array, their index corresponds to their*
 * id number.                                                                 *
 * Similiarly, cable modem leaf nodes are all stored in an array, and their   *
 * index corresponds to their id number, offsetted by the number of university*
 * nodes.
 ******************************************************************************/

/**********
 * MACROS *
 **********/

#define SET_LINK(link, linktype, sender, receiver) \
	do { \
		link->type = linktype; \
		link->bandwidth = 0; \
		link->send_node = sender; \
		link->recv_node = receiver; \
		link->connections.clear(); \
		link->conns_copy.clear(); \
		link++; \
	} while (0)

#define NUM_CLIENTS		(_num_univ_nodes + _num_cable_leaf_nodes)
#define NUM_UNIV_UNIV_LINKS	(_num_univ_nodes - 1)
#define ID_IN_AS(id, as) \
	((id) >= (as)->low_client_id && (id) <= (as)->hi_client_id)
#define ID_IN_HEADEND(id, he) \
	((id) >= (he)->low_client_id && (id) <= (he)->hi_client_id)


/*********
 * TYPES *
 *********/

struct conn;
struct link;
struct bytes_transfer;
	
/* typedef'd STL types */
typedef pair<unsigned int, unsigned int> conn_pair;
typedef map<conn_pair, struct conn*> conn_map;
typedef vector<struct link*> link_vector;
typedef set<struct conn*> conn_set;
typedef vector<struct bytes_transfer> transfer_vector;

/* bytes transferred from a sender to a receiver */
struct bytes_transfer {
	unsigned int sender_id;		/* sender of the bytes */
	unsigned int receiver_id;	/* receiver of the bytes */
	unsigned int num_bytes;		/* number of bytes transferred */
};

/* connection type */
struct conn {
	unsigned int tick_update;	/* last time did data update */
	unsigned int sender_id;		/* client id of the sender */
	unsigned int receiver_id;	/* client id of the receiver */
	unsigned char closeness;	/* closeness of sender<->receiver */
#ifdef _SOLARIS_ /* g++ solaris can't handle packed structs???  pathetic! */
};
#else
} __attribute__ ((__packed__));
#endif

/* connection bandwidth pair */
struct conn_bw {
	struct conn *conn;		/* data transfer connection */
	float bandwidth;		/* bandwidth allocated to that conn */
};

/* link types */
enum {
	UNIV_UNIV = 0,			/* UNIVERSITY TO UNIVERSITY LINK */
	UNIV_VIRT,			/* UNIVERSITY TO VIRTUAL LINK */
	VIRT_UNIV,			/* VIRTUAL TO UNIVERSITY LINK */
	UNIV_AS,			/* UNIVERSITY TO AS LINK */
	AS_UNIV,			/* AS TO UNIVERSITY LINK */
	AS_AS,				/* AS TO AS LINK */
	AS_VIRT,			/* AS TO VIRTUAL LINK */
	VIRT_AS,			/* VIRTUAL TO AS LINK */
	HEADEND_HEADEND,		/* HEADEND TO HEADEND LINK */
	HEADEND_VIRT,			/* HEADEND TO VIRTUAL NODE LINK */
	VIRT_HEADEND,			/* VIRTUAL NODE TO HEADEND LINK */
	VIRT_LEAF,			/* VIRTUAL NODE TO CABLE LEAF LINK */
	LEAF_VIRT			/* CABLE LEAF TO VIRTUAL NODE LINK */
};

/* link (graph edge) type */
struct link {
	float bandwidth;		/* available bandwidth on link */
	conn_set connections;		/* set of connections on this link */
	conn_set conns_copy;		/* copy of the set of connections */
	void *send_node;		/* sending node */
	void *recv_node;		/* receiving node */
	unsigned char type;		/* link type */
#ifdef _SOLARIS_
};
#else
} __attribute__ ((__packed__));
#endif

/* graph node types */

struct as_virt {
	struct link *as_link;		/* link to virtual node's AS node */
	struct link *univ_links;	/* _num_univ_nodes of
					   university links */
	struct link *gateway_links;	/* _num_as-1 of gateway links */
	double outgoing_bw_percent;	/* percent of outgoing bw utilization */
	double incoming_bw_percent;	/* percent of incoming bw utilization */
};

struct as {
	struct as_virt virt_node;	/* AS's virtual node */
	struct link *virt_link;		/* AS's virtual shared external link */
	unsigned int num_nodes;		/* number of headend offices in AS */
	unsigned int low_client_id;	/* lowest client id in AS */
	unsigned int hi_client_id;	/* highest client id in AS */
	struct headend_node *gateway;	/* AS's gateway node */
	bt_subtracker subtracker;	/* the as's subtracker */
};

struct univ_virt {
	struct link *univ_link;		/* link to virtual node's univ node */
	struct link *univ_links;	/* _num_univ_nodes-1 of
					   university links */
	struct link *gateway_links;	/* _num_as-1 of gateway links */
	double outgoing_bw_percent;	/* percent of outgoing bw utilization */
	double incoming_bw_percent;	/* percent of incoming bw utilization */
};

struct univ_node {
	struct univ_virt virt_node;	/* univ's virtual node */
	struct link *virt_link;		/* link to virtual node */

	/* added by Ruchir on 10/01/05	*/
	bt_subtracker* univ_subtracker; 	/* pointer to the university AS subtracker */

};

struct cable_virt {
	struct link *headend_link;	/* link to the headend */
	struct link *cable_leaf_links;	/* links to the cable modem users */
	double outgoing_bw_percent;	/* percent of outgoing bw utilization */
	double incoming_bw_percent;	/* percent of incoming bw utilization */
};

struct headend_node {
	struct cable_virt virt_node;	/* headend's virtual node */
	struct as *my_as;		/* headend's as */
	unsigned int low_client_id;	/* lowest client id contained in
					   this headend node */
	unsigned int hi_client_id;	/* highest client id contained in
					   this headend node */
	struct link *virt_link;		/* virtual link shared by cable modem
					   neighbors */
	struct link *headend_links;	/* link to all other headend nodes */
};

struct cable_leaf_node {
	struct link *virt_link;		/* the virtual link */
	double outgoing_bw_percent;	/* percent of outgoing bw utilization */
	double incoming_bw_percent;	/* percent of incoming bw utilization */
};

/* enum for links' connection manipulation flag */
enum {
	ADD = 0,			/* add connection to link */
	REMOVE				/* remove connection from link */
};

/* bandwidth comparator */
struct more_bandwidth {
	bool operator()(struct link *x, struct link *y) {
		assert(x && y);
		float xbw = (float)(x->bandwidth) /
			(float)(x->conns_copy.size());
		float ybw = (float)(y->bandwidth) /
			(float)(y->conns_copy.size());
		return xbw > ybw;
	}
};



/*************
 * CONSTANTS *
 *************/


static const unsigned int MAX_CHARS_PER_LINE = 80;
static const unsigned int NUM_LINK_TYPES = 13;


/********************
 * GLOBAL VARIABLES *
 ********************/

/* topology tracking */
unsigned long long g_closeness[DEGREES_OF_CLOSENESS];	/* conn closeness */

/* upload bandwidth tracking */
static unsigned long long _bw_log_events = 0;		/* num of events */

/* topology file configuration variables */

static int _seed_is_cable = 0;		/* 0 = seed is a university node
        				   non-zero = seed is cable node */
static int _arrivals_are_random = 0;	/* 0 = flash crowd, all join at once */
static int _departures_are_random = 0;	/* 0 = peer departs immediately after
					   completing download */
static unsigned int _num_pieces = 0;	/* number of pieces in file */
static unsigned int _tracker_type = NORMAL_TRACKER;  /* tracker type */
static unsigned int _num_external_peers = 0; /* number of non local peers
						returned by subtracker */
static float _bandwidth[NUM_LINK_TYPES];/* link bandwidths */

/* topology node variables */

static unsigned int _num_as = 0;	/* the number of ASes in the topology */
static struct as *_as = NULL;		/* array of ASes in the topology, 
					   element 0 is the university AS */
static unsigned int _num_univ_nodes = 0;/* the number of university nodes */
static struct univ_node *_univ_node = NULL;	/* the university nodes */
static unsigned int _num_headend_nodes = 0;	/* number of headend offices */
static struct headend_node *_headend_node = NULL;	/* the headend nodes */
static unsigned int _num_cable_leaf_nodes = 0;	/* number of cable leaf nodes */
static struct cable_leaf_node *_cable_leaf_node = NULL;	/* the leaf nodes */
static unsigned int _num_links = 0;	/* number of links */
static struct link *_links = NULL;	/* array of links */

/* bt_peer connections */

static struct memory_pool _conn_pool;	/* memory pool of connections */
static conn_map _connection_map;	/* hashtable of all the connections */

/* bt_peer's */

static bt_peer *_peers = NULL;		/* array of bt_peers */

/* bt_tracker */

static bt_tracker _tracker;		/* the torrent tracker */

/* univ nodes subtracker */
static bt_subtracker* univ_sub;


/* data transfer calculation variables */

static link_vector _link_vec;		/* vector of links for data calc */
static transfer_vector _xfer_vec;	/* vector of bytes transferred */


/***********************
 * FUNCTION PROTOTYPES *
 ***********************/


/**
 * Reads an unsigned int from the topology file.  Sanity checks to make sure
 * it's a number.
 * @param fp		the topology file pointer
 * @return -1 on error, else the integer value of the number in the file's line
 */
static int fread_topo_uint(FILE *fp);

/**
 * Reads a float from the topology file.  Sanity checks to make sure it's a
 * number.
 * @param fp		the topology file pointer
 * @return -1.0 on error, else the float value of the number in the file's line
 */
static float fread_topo_float(FILE *fp);

/**
 * Reads a line from the topology file.  Sanity checks the line.  Keeps state
 * like the line number.  Automatically reads another line if the first is 
 * commented.
 * @param fp		the topology file pointer
 * @param buf		the buffer to write the file data to
 * @param len		the buffer length
 * @return 0 on success, non-zero on error
 */
static int fread_topo_line(FILE *fp, char buf[], unsigned int len);

/**
 * Reads from a topology file the lines corresponding to a university as.
 * Creates the necessary university nodes.
 * @param fp		the topology file pointer
 * @return 0 on success, non-zero on error
 */
static int fread_topo_univ_as(FILE *fp);

/**
 * Reads from a topology file the lines corresponding to a cable modem as.
 * Creates the necessary nodes.  Sets the client id fields.
 * @param as		the AS node
 * @param fp		the topology file pointer
 * @return 0 on success, non-zero on error
 */
static int fread_topo_as(struct as *as, FILE *fp);

/**
 * Allocates memory for the links.  Goes through all the nodes and counts how 
 * many are necessary.
 */
static void alloc_links();

/**
 * Goes through all the nodes and sets their link pointers and assigns cable 
 * leaf nodes to virtual nodes.
 */
static void init_all_nodes();

/**
 * Sets the links for an university node.
 * @param univ		the university node
 * @param curr_link	the current link
 * @return the new current link pointer
 */
static struct link *init_univ(struct univ_node *univ, struct link *curr_link);

/**
 * Sets the links for an AS.  Recurses into setting links for children too.
 * @param as_node	the AS node
 * @param curr_link	the current link
 * @param gateway	the as's gateway headend office node
 * @return the new current link pointer
 */
static struct link *init_as(struct as *as_node,
		struct link *curr_link,
		struct headend_node *gateway);

/**
 * Sets the links for a headend office.  Recurses into setting links for 
 * the children nodes as well (virtual node + leaf nodes).
 * @param he		the headend office node
 * @param as_node	the AS node
 * @param curr_link	the current link
 * @return the new current link pointer
 */
static struct link *init_headend(struct headend_node *he,
		struct as *as_node,
		struct link *curr_link);

/**
 * Obtains a connection from a connection pool.  Marks it as used.
 * @return a connection structure, or NULL if out of available connections
 */
static __inline__ struct conn *conn_alloc();

/**
 * Returns an allocated connection back to the connection pool.  Marks it 
 * as unused.
 * @param conn		the connection to mark as unused
 */
static __inline__ void conn_free(struct conn *conn);

/**
 * Adds a connection to the connection map.
 * @param conn		the connection to add
 */
static __inline__ void map_add_conn(struct conn *conn);

/**
 * Removes a connection from the connection map.
 * @param conn		the connection to remove
 */
static __inline__ void map_remove_conn(struct conn *conn);

/**
 * Converts a bt_peer* pointer to a client id.
 * @param peer		pointer to the bt_peer
 * @return the client id
 */
static __inline__ unsigned int get_peer_id(bt_peer *peer);

/**
 * Iterates through the links from sender to receiver.  Either adds or 
 * removes the connection along the link.
 * @param sender_id	client id of the sending peer
 * @param receiver_id	client id of the receiving peer
 * @param travel_link	function to call per link traveled
 * @param link_args	the arguments for travel_link
 */
static __inline__ void link_traveler(unsigned int sender_id,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args);

/**
 * Iterates through the links from sender to receiver, where the sender is a 
 * university node.  Either adds or removes the connection along the link.
 * @param sender   	the sending university node
 * @param receiver_id	client id of the receiving peer
 * @param travel_link	function to call per link traveled
 * @param link_args	the arguments for travel_link
 */
static void univ_link_traveler(struct univ_node *sender,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args);

/**
 * Iterates through the links from sender to receiver, where the sender is a
 * cable leaf node.  Either adds or removes the connection along the link.
 * @param sender   	the sending cable leaf node
 * @param receiver_id	client id of the receiving peer
 * @param travel_link	function to call per link traveled
 * @param link_args	the arguments for travel_link
 */
static void cable_leaf_link_traveler(struct cable_leaf_node *sender,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args);

/**
 * Iterates through the links from an AS gateway to the receiver, where you
 * know the receiver is within the AS.  Either adds or removes the
 * connection along the link.
 * @param as		the AS
 * @param receiver_id	client id of the receiving peer
 * @param travel_link	function to call per link traveled
 * @param link_args	the arguments for travel_link
 */
static void as_link_traveler(struct as *as,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args);

/**
 * Iterates through the links from an AS headend to the receiver, where you
 * know the receiver is within the headend office.  Either adds or removes the
 * connection along the link.
 * @param as		the AS
 * @param receiver_id	client id of the receiving peer
 * @param travel_link	function to call per link traveled
 * @param link_args	the arguments for travel_link
 */
static __inline__ void headend_link_traveler(struct headend_node *he,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args);

/**
 * Adds a data connection to a link.
 * @param link		the link to add a connection to
 * @param args		the connection to add
 */
static void add_conn_to_link(struct link *link, void *args);

/**
 * Removes a data connection from a link.
 * @param link		the link to remove a connection from
 * @param args		the connection to remove
 */
static void remove_conn_from_link(struct link *link, void *args);

/**
 * Returns whether or not a client is a university node.  If not, it's a 
 * cable leaf node.
 * @param client_id	id of the client being identified
 * @return 0 if cable leaf node, non-zero if university node
 */
static __inline__ int is_univ_node(unsigned int client_id);

/**
 * Returns the university node associated with the client id.
 * @param client_id	id of the university node
 * @return a pointer to the university node
 */
static __inline__ struct univ_node *get_univ_node(unsigned int client_id);

/**
 * Returns the cable leaf node associated with the client id.
 * @param client_id	id of the cable leaf node
 * @return a pointer to the cable leaf node
 */
static __inline__ struct cable_leaf_node *get_leaf_node(unsigned int client_id);

/**
 * Converts a university node pointer to a client id.
 * @param node		the university node
 * @return the client id of the node
 */
static __inline__ unsigned int univ_get_client_id(struct univ_node *univ);

/**
 * Converts a cable leaf node pointer to a client id.
 * @param node		the cable leaf node
 * @return the client id of the node
 */
static __inline__ unsigned int leaf_get_client_id(struct cable_leaf_node *leaf);

/**
 * Checks to see if a link matches the description.
 * @param link		the link we're trying to match
 * @param send_node	the sending end of the link
 * @param recv_node	the receiving end of the link
 * @return non-zero if it matches, 0 if it doesn't match
 */
static __inline__ unsigned int link_match(struct link *link,
		void *send_node,
		void *recv_node);

/**
 * Removes the connection of the link and deducts its bandwidth usage from
 * the available bandwidth
 * @param link		the link we're operating on
 * @param args		the connection and its bandwidth usage
 */
static void adjust_bandwidth(struct link *link, void *args);

/**
 * Updates the bandwidth usage tracking.
 */
static void update_bw_usage();


/************************
 * FUNCTION DEFINITIONS *
 ************************/


bt_tracker *get_tracker(bt_peer *peer)
{
	assert(peer);

	unsigned int peer_id = peer - _peers;

	/* modified by Ruchir on June 28'05 to incorporate the SUB_LIMIT_TRACKER*/
	if ((_tracker_type != SUB_TRACKER && _tracker_type != SUB_LIMIT_TRACKER) || is_univ_node(peer_id)) {
		return &_tracker;
	}

	struct cable_leaf_node *cable_leaf = get_leaf_node(peer_id);
	struct cable_virt *cable_virt = (struct cable_virt*)
		cable_leaf->virt_link->recv_node;
	struct headend_node *he = (struct headend_node*)
		cable_virt->headend_link->recv_node;
	struct as *as = he->my_as;

	return &as->subtracker;
}

unsigned int get_tracker_type()
{
	return _tracker_type;
}

int get_seed_is_cable()
{
	return _seed_is_cable;
}

int get_arrivals_are_random()
{
	return _arrivals_are_random;
}

int get_departures_are_random()
{
	return _departures_are_random;
}

unsigned int get_num_pieces()
{
	return _num_pieces;
}

unsigned int get_num_univ_nodes()
{
	return _num_univ_nodes;
}

int init_topology(const char *topo_file)
{
	assert(topo_file);

	for (unsigned int i = 0; i < DEGREES_OF_CLOSENESS; i++) {
		g_closeness[i] = 0;
	}

	int ret = 0;
	FILE *fp = fopen(topo_file, "r");
	_EXPECT_NOTNULL(fp, "Error: topology file inaccessible.\n");
	debug_msg("Opened topology file: %s\n", topo_file);

	/* read the settings */
	_arrivals_are_random = fread_topo_uint(fp);
	debug_msg("Random arrivals: %d\n", _arrivals_are_random);
	_departures_are_random = fread_topo_uint(fp);
	debug_msg("Random departures: %d\n", _departures_are_random);
	_num_pieces = fread_topo_uint(fp);
	debug_msg("Number of pieces in file: %d\n", _num_pieces);
	_seed_is_cable = fread_topo_uint(fp);
	debug_msg("Cable seed: %d\n", _seed_is_cable);
	_tracker_type = fread_topo_uint(fp);
	switch (_tracker_type) {
		case NORMAL_TRACKER:
			debug_msg("Using normal tracker.\n");
			break;
		case SUB_TRACKER:
			debug_msg("Using subtrackers.\n");
			_num_external_peers = fread_topo_uint(fp);
			debug_msg("\tNumber of non-local peers: %u\n",
					_num_external_peers);
			break;
		/* added by Ruchir on June 28'05 -- 
			this type of tracker controls the number of connections
			outside of an AS on a per-AS basis and not on per node basis*/
		case SUB_LIMIT_TRACKER:
			debug_msg("Using as limit subtrackers.\n");
			_num_external_peers = fread_topo_uint(fp);
			debug_msg("\tNumber of non-local peers per AS: %u\n",
					_num_external_peers);
			break;

		case UNIV_TRACKER:
			debug_msg("Using university biased tracker.\n");
			_num_external_peers = fread_topo_uint(fp);
			debug_msg("\tNumber of non-univ peers: %u\n",
					_num_external_peers);
			break;
	}

	for (unsigned int i = 0; i < NUM_LINK_TYPES; i++) {
		_bandwidth[i] = fread_topo_float(fp);
		debug_msg("Bandwidth[%d] = %f\n", i, _bandwidth[i]);
	}

	/* added by Ruchir on 10/01/05 */

	/* create the university subtracker */
	univ_sub = new bt_subtracker();
	univ_sub->set_tracker(&_tracker);
	/* now just set the univ nodes' pointers to this pointer*/


	/* read the university nodes */
	EXPECT_ZERO(fread_topo_univ_as(fp));
	_EXPECT_TRUE(_seed_is_cable || _num_univ_nodes,
			"The seed can't be a university node if there aren't any university nodes.\n");

	if(_num_univ_nodes > 0)
	{
		assert(_univ_node);
		for(unsigned int i = 0; i < _num_univ_nodes; i++)
		{
			_univ_node[i].univ_subtracker = univ_sub;	
		}
	}

	/* read the number of ASes */
	_num_as = fread_topo_uint(fp);
	debug_msg("Number of ASes to read: %d\n", _num_as);

	_as = new struct as[_num_as];

	for (unsigned int i = 0; i < _num_as; i++) {
		debug_msg("Reading AS %d from topology file.\n", i);
		EXPECT_ZERO(fread_topo_as(&_as[i], fp));
		_as[i].subtracker.set_tracker(&_tracker);
	}

	_cable_leaf_node = (struct cable_leaf_node*)_malloc(
			_num_cable_leaf_nodes * sizeof(*_cable_leaf_node));

	alloc_links();
	init_all_nodes();

	_peers = new bt_peer[NUM_CLIENTS];
	assert(_peers);
	init_memory_pool(&_conn_pool,
			NUM_CLIENTS * MAX_UPLOADS,
			sizeof(struct conn));
	_connection_map.clear();
	_bw_log_events = 0;

	_link_vec.clear();
	_link_vec.reserve(_num_links);
	_xfer_vec.clear();
	_xfer_vec.reserve(NUM_CLIENTS * MAX_UPLOADS);

_cleanup:
	SAFE_CLOSE(fp);

	if (ret) {
		SAFE_DELETE(_as);
		SAFE_DELETE(_peers);
		SAFE_FREE(_univ_node);
		SAFE_FREE(_headend_node);
		SAFE_FREE(_cable_leaf_node);
	}

	return ret;
}

static int fread_topo_uint(FILE *fp)
{
	assert(fp);
	int ret = 0;
	char buf[MAX_CHARS_PER_LINE + 2]; /* add space for \n\0 */
	int num = -1;

	EXPECT_ZERO(fread_topo_line(fp, buf, MAX_CHARS_PER_LINE + 2));

	num = strtol(buf, NULL, 10);
	_EXPECT_FALSE(num < 0 || (!num && errno == EINVAL),
			"Error: %s is not a valid number.\n",
			buf);
	ret = num;

_cleanup:

	return ret;
}

static float fread_topo_float(FILE *fp)
{
	assert(fp);
	float ret = 0;
	char buf[MAX_CHARS_PER_LINE + 2]; /* add space for \n\0 */
	float num = -1.0;

	EXPECT_ZERO(fread_topo_line(fp, buf, MAX_CHARS_PER_LINE + 2));

	num = atof(buf);
	_EXPECT_FALSE(num < 0 || (!num && errno == EINVAL),
			"Error: %s is not a valid number.\n",
			buf);
	ret = num;

_cleanup:

	return ret;
}

static int fread_topo_line(FILE *fp, char buf[], unsigned int len)
{
	static int linenum = 1;
	int ret = 0;
	char *c = NULL;

	assert(fp && buf);
	
	EXPECT_NOTNULL(fgets(buf, MAX_CHARS_PER_LINE + 2, fp));
	
	c = strchr(buf, '\n');
	_EXPECT_TRUE(c || feof(fp), "Error: line %d exceeds %d characters.\n",
				linenum,
				MAX_CHARS_PER_LINE);

	if (buf[0] == '\n' || buf[0] == '#') {
		linenum++;
		return fread_topo_line(fp, buf, len);
	}

	buf[strlen(buf)] = 0;

	linenum++;

_cleanup:

	return ret;
}

static int fread_topo_univ_as(FILE *fp)
{
	assert(fp);
	int ret = 0;
	_num_univ_nodes = fread_topo_uint(fp);
	
	debug_msg("Number of university nodes: %d.\n", _num_univ_nodes);

	if (_num_univ_nodes > 0) {
		_univ_node = (struct univ_node*)_malloc(sizeof(*_univ_node) *
			_num_univ_nodes);
	}

	return ret;
}

static int fread_topo_as(struct as *as, FILE *fp)
{
	assert(as && fp);
	int ret = 0;

	/* read configuration for the AS */
	unsigned int gateway_node_idx = _num_headend_nodes;
	as->low_client_id = _num_univ_nodes + _num_cable_leaf_nodes;
	as->num_nodes = fread_topo_uint(fp);
	EXPECT_TRUE(as->num_nodes > 0);
	debug_msg("\tNumber of headend offices in AS: %d.\n", as->num_nodes);

	_num_headend_nodes += as->num_nodes;
	_headend_node = (struct headend_node*)_realloc(_headend_node,
			_num_headend_nodes * sizeof(*_headend_node));
	as->gateway = &_headend_node[gateway_node_idx];

	/* read configuration for each headend office */
	for (unsigned int i = gateway_node_idx; i < _num_headend_nodes; i++) {
		_headend_node[i].low_client_id =
			_num_univ_nodes + _num_cable_leaf_nodes;
		int fanout = fread_topo_uint(fp);
		EXPECT_TRUE(fanout > 0);
		debug_msg("\t\tOffice %d has fanout: %d\n",
				i - gateway_node_idx,
				fanout);
		_num_cable_leaf_nodes += fanout;
		_headend_node[i].hi_client_id =
			_headend_node[i].low_client_id + fanout - 1;
		_headend_node[i].my_as = as;
		_headend_node[i].virt_link = NULL;
		_headend_node[i].headend_links = NULL;
		_headend_node[i].virt_node.outgoing_bw_percent = 0;
		_headend_node[i].virt_node.incoming_bw_percent = 0;
	}
	as->hi_client_id = _num_univ_nodes + _num_cable_leaf_nodes - 1;
	as->virt_node.outgoing_bw_percent = 0;
	as->virt_node.incoming_bw_percent = 0;

_cleanup:

	if (ret) {
		DEBUG_ERROR("Error processing AS data.\n");
	}

	return ret;
}

unsigned int get_num_peers()
{
	return NUM_CLIENTS;
}

bt_peer *get_peers()
{
	assert(_peers);
	return _peers;
}

static void alloc_links()
{
	_num_links = 0;

	/* each university wants a link to every other university */
	_num_links += _num_univ_nodes * (_num_univ_nodes - 1);

	/* each university wants a link to every AS gateway */
	_num_links += _num_univ_nodes * _num_as;

	/* each university has a 2-way virtual link */
	_num_links += 2 * _num_univ_nodes;

	/* each AS gateway wants a link to every university */
	_num_links += _num_as * _num_univ_nodes;

	/* each AS gateway wants a link to every other AS gateway */
	_num_links += _num_as * (_num_as - 1);

	/* each AS gateway has a 2-way virtual link */
	_num_links += 2 * _num_as;

	struct headend_node *he = &_headend_node[0];
	for (unsigned int i = 0; i < _num_as; i++) {
		/* each headend wants a link to every other headend in AS */
		_num_links += _as[i].num_nodes * (_as[i].num_nodes - 1);

		/* each headend has a link to and from the virtual node */
		_num_links += 2 * _as[i].num_nodes;

		/* each virtual node has a link to and from each leaf */

		for (unsigned int j = 0; j < _as[i].num_nodes; j++, he++) {
			assert(he->hi_client_id >= he->low_client_id);
			_num_links += 2 *
				(1 + he->hi_client_id - he->low_client_id);
		}
	}

	_links = new struct link[_num_links];
}

static void init_all_nodes()
{
	assert(_links);
	struct link *curr_link = _links;

	/* set all the university nodes' links */
	for (unsigned int i = 0; i < _num_univ_nodes; i++) {
		curr_link = init_univ(&_univ_node[i], curr_link);
	}

	/* set all the as nodes' and children nodes' links */
	struct headend_node *gateway = &_headend_node[0];
	for (unsigned int i = 0; i < _num_as; i++) {
		curr_link = init_as(&_as[i], curr_link, gateway);
		gateway += _as[i].num_nodes;
	}

	assert(_num_links == (unsigned int)(curr_link - _links));
}

static struct link *init_univ(struct univ_node *univ, struct link *curr_link)
{
	assert(univ && curr_link);

	univ->virt_link = curr_link;
	SET_LINK(curr_link, UNIV_VIRT, univ, &univ->virt_node);

	struct univ_virt *virt_node = &univ->virt_node;
	virt_node->univ_link = curr_link;
	SET_LINK(curr_link, VIRT_UNIV, virt_node, univ);
	virt_node->univ_links = curr_link;

	for (unsigned int i = 0; i < _num_univ_nodes; i++) {
		if (univ == &_univ_node[i]) {
			continue;
		}
		SET_LINK(curr_link,
				UNIV_UNIV,
				virt_node,
				&_univ_node[i].virt_node);
	}

	virt_node->gateway_links = curr_link;
	
	for (unsigned int i = 0; i < _num_as; i++) {
		SET_LINK(curr_link, UNIV_AS, virt_node, &_as[i].virt_node);
	}

	return curr_link;
}

static struct link *init_as(struct as *as_node,
		struct link *curr_link,
		struct headend_node *gateway)
{
	assert(as_node && curr_link && gateway);

	as_node->gateway = gateway;

	/* set the AS's virtual shared links */
	struct as_virt *virt_node = &as_node->virt_node;
	as_node->virt_link = curr_link;
	SET_LINK(curr_link, AS_VIRT, as_node, virt_node);

	virt_node->as_link = curr_link;
	SET_LINK(curr_link, VIRT_AS, virt_node, as_node);

	/* set the AS's links to universities */

	virt_node->univ_links = curr_link;

	for (unsigned int i = 0; i < _num_univ_nodes; i++) {
		SET_LINK(curr_link,
				AS_UNIV,
				virt_node,
				&_univ_node[i].virt_node);
	}

	/* set the AS's links to other AS's */

	virt_node->gateway_links = curr_link;

	for (unsigned int i = 0; i < _num_as; i++) {
		if (as_node == &_as[i]) {
			continue;
		}
		SET_LINK(curr_link, AS_AS, virt_node, &_as[i].virt_node);
	}

	/* set each headend node's links */
	struct headend_node *he = as_node->gateway;
	for (unsigned int i = 0; i < as_node->num_nodes; i++, he++) {
		curr_link = init_headend(he, as_node, curr_link);
	}

	return curr_link;
}

static struct link *init_headend(struct headend_node *he,
		struct as *as_node,
		struct link *curr_link)
{
	assert(he && as_node && curr_link && as_node->gateway);

	struct headend_node *he_peer = as_node->gateway;
	he->headend_links = curr_link;

	for (unsigned int i = 0; i < as_node->num_nodes; i++, he_peer++) {
		if (he == he_peer) {
			continue;
		}

		SET_LINK(curr_link, HEADEND_HEADEND, he, he_peer);
	}

	assert(he != as_node->gateway || he != he->headend_links[0].recv_node);

	he->virt_link = curr_link;
	SET_LINK(curr_link, HEADEND_VIRT, he, &he->virt_node);
	he->virt_node.headend_link = curr_link;
	SET_LINK(curr_link, VIRT_HEADEND, &he->virt_node, he);

	he->virt_node.cable_leaf_links = curr_link;

	unsigned int low_leaf_idx = he->low_client_id - _num_univ_nodes;
	unsigned int max_leaf_idx = he->hi_client_id - _num_univ_nodes;

	for (unsigned int i = low_leaf_idx; i <= max_leaf_idx; i++) {
		SET_LINK(curr_link,
				VIRT_LEAF,
				&he->virt_node,
				&_cable_leaf_node[i]);
	}

	for (unsigned int i = low_leaf_idx; i <= max_leaf_idx; i++) {
		_cable_leaf_node[i].virt_link = curr_link;
		SET_LINK(curr_link,
				LEAF_VIRT,
				&_cable_leaf_node[i],
				&he->virt_node);
	}

	return curr_link;
}

void add_data_transfer(bt_peer *sender,
		bt_peer *receiver)
{
	assert(sender && receiver);

	unsigned int sender_id = get_peer_id(sender);
	unsigned int receiver_id = get_peer_id(receiver);

	conn_map::const_iterator it =
		_connection_map.find(conn_pair(sender_id, receiver_id));
	assert(it == _connection_map.end());

	struct conn *data_conn = conn_alloc();
	assert(data_conn);

	data_conn->tick_update = g_tick;
	data_conn->sender_id = sender_id;
	data_conn->receiver_id = receiver_id;
	if (is_univ_node(sender_id)) {
		if (is_univ_node(receiver_id)) {
			data_conn->closeness = INTRA_UNIV;
		} else {
			data_conn->closeness = INTER;
		}
	} else {
		if (is_univ_node(receiver_id)) {
			data_conn->closeness = INTER;
		} else {
			struct cable_leaf_node *send_node =
				get_leaf_node(sender_id);
			struct cable_virt *virt = (struct cable_virt*)
				send_node->virt_link->recv_node;
			struct headend_node *he = (struct headend_node*)
				virt->headend_link->recv_node;
			if (receiver_id >= he->low_client_id &&
					receiver_id <= he->hi_client_id) {
				data_conn->closeness = INTRA_HEADEND;
			} else {
				struct as *as = he->my_as;
				if (receiver_id >= as->low_client_id &&
					receiver_id <= as->hi_client_id) {
					data_conn->closeness = INTRA_AS;
				} else {
					data_conn->closeness = INTER;
				}
			}
		}
	}

	map_add_conn(data_conn);

	link_traveler(sender_id, receiver_id, add_conn_to_link, data_conn);

	DEBUG_MSG("Added data connection from peer %d to peer %d to graph.\n",
		sender_id,
		receiver_id);
}

void remove_data_transfer(bt_peer *sender, bt_peer *receiver)
{
	assert(sender && receiver);

	unsigned int sender_id = get_peer_id(sender);
	unsigned int receiver_id = get_peer_id(receiver);

	conn_map::const_iterator it =
		_connection_map.find(conn_pair(sender_id, receiver_id));
	assert(it != _connection_map.end());
	struct conn *data_conn = it->second;
	assert(data_conn);

	link_traveler(sender_id, receiver_id, remove_conn_from_link, data_conn);
	map_remove_conn(data_conn);
	conn_free(data_conn);

	DEBUG_MSG("Removed data connection from peer %d"
		" to peer %d from graph.\n",
		sender_id,
		receiver_id);
}

static __inline__ struct conn *conn_alloc()
{
	struct conn *data_conn =
		(struct conn*)pool_object_alloc(&_conn_pool);
	assert(data_conn);
	return data_conn;
}

static __inline__ void conn_free(struct conn *conn)
{
	pool_object_free(&_conn_pool, conn);
}

static __inline__ void map_add_conn(struct conn *conn)
{
	assert(conn);
	
	pair<conn_pair, struct conn*> conn_value(conn_pair(conn->sender_id,
				conn->receiver_id), conn);
	_connection_map.insert(conn_value);
}

static __inline__ void map_remove_conn(struct conn *conn)
{
	assert(conn);
	
	_connection_map.erase(conn_pair(conn->sender_id, conn->receiver_id));
}

static __inline__ unsigned int get_peer_id(bt_peer *peer)
{
	assert(peer && peer >= _peers);

	return peer - _peers;
}

static __inline__ void link_traveler(unsigned int sender_id,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args)
{
	if (is_univ_node(sender_id)) {
		struct univ_node *univ = get_univ_node(sender_id);
		univ_link_traveler(univ,
				receiver_id,
				travel_link,
				args);
	} else {
		struct cable_leaf_node *leaf = get_leaf_node(sender_id);
		cable_leaf_link_traveler(leaf,
				receiver_id,
				travel_link,
				args);
	}
}

static void univ_link_traveler(struct univ_node *sender,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args)
{
	assert(sender);

	/* path: univ->univ_virt->univ?as? */

	struct link *link = sender->virt_link;
	assert(link->send_node == sender &&
			link->recv_node == &sender->virt_node);
	travel_link(link, args);

	/* path: univ_virt->univ?as? */

	struct univ_virt *virt_node = &sender->virt_node;

	if (is_univ_node(receiver_id)) {
		/* path: univ_virt->peer_univ_virt->peer_univ */
		struct univ_node *receiver = get_univ_node(receiver_id);
		assert(receiver);
		struct univ_virt *recv_virt = &receiver->virt_node;
		for (unsigned int i = 0; i < NUM_UNIV_UNIV_LINKS; i++) {
			link = &virt_node->univ_links[i];
			if (link_match(link, sender, recv_virt)) {
				travel_link(link, args);
				/* path: peer_univ_virt->peer_univ */
				link = recv_virt->univ_link;
				assert(link->send_node == recv_virt &&
					link->recv_node == receiver);
				travel_link(link, args);
				/* path complete */
				return;
			}
		}
		/* should never reach here */
		BUG();
	} else {
		/* first, we need to figure out which AS has the receiver id */
		struct as *as = NULL;
		link = NULL;

		/* path: univ_virt->as_virt->as->inside_as */

		for (unsigned int i = 0; i < _num_as; i++) {
			if (ID_IN_AS(receiver_id, &_as[i])) {
				as = &_as[i];
				link = &virt_node->gateway_links[i];
				break;
			}
		}
		assert(as);
		assert(link->send_node == virt_node &&
			link->recv_node == &as->virt_node);
		travel_link(link, args);

		/* path: as_virt->as->inside_as */

		as_link_traveler(as,
				receiver_id,
				travel_link,
				args);
	}
}

static void cable_leaf_link_traveler(struct cable_leaf_node *sender,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args)
{
	assert(sender && sender->virt_link);

	/* first, we try to do intra AS search */

	/* path: cable_modem->cable_virt->? */

	/* search the shared link neighbors */
	struct link *link = sender->virt_link;
	assert(link->send_node == sender);
	travel_link(link, args);
	/* path: cable_virt->? */
	struct cable_virt *virt_node = (struct cable_virt*)link->recv_node;
	assert(virt_node);
	struct headend_node *he = (struct headend_node*)
		virt_node->headend_link->recv_node;
	assert(he);
	if (ID_IN_HEADEND(receiver_id, he)) {
		/*
		link = virt_node->headend_link;
		travel_link(link, args);
		link = he->virt_link;
		travel_link(link, args);
		*/
		/* found the neighbor, done! */
		/* path: cable_virt->cable_modem */
		unsigned int leaf_offset = receiver_id - he->low_client_id;
		link = &virt_node->cable_leaf_links[leaf_offset];
		assert(link->send_node == virt_node &&
				link->recv_node == get_leaf_node(receiver_id));
		travel_link(link, args);
		/* path complete */
		return;
	}

	/* wasn't a direct neighbor, check to see if it's within the AS */
	/* path: cable_virt->headend->?*/
	link = virt_node->headend_link;
	travel_link(link, args);
	/* path: headend->?*/
	struct as *as = he->my_as;
	if (ID_IN_AS(receiver_id, as)) {
		/* path: headend->headend->cable_virt->cable_modem */
		struct headend_node *peer_he = as->gateway;
		unsigned int i;
		for (i = 0; i < as->num_nodes; i++) {
			if (ID_IN_HEADEND(receiver_id, &peer_he[i])) {
				peer_he = &peer_he[i];
				assert(he != peer_he);
				if (he < peer_he) {
					link = &he->headend_links[i-1];
				} else {
					link = &he->headend_links[i];
				}
				assert(link->send_node == he &&
						link->recv_node == peer_he);
				travel_link(link, args);
				break;
			}
		}
		assert(i < as->num_nodes);

		/* path: headend->cable_virt->cable_modem */

		/* find the peer within the headend */
		headend_link_traveler(peer_he,
				receiver_id,
				travel_link,
				args);
	} else {
		/* receiver is in another AS, so travel link to gateway if
		   is not gateway */
		if (he != as->gateway) {
			/* path: headend->gateway->as_virt->outside */
			link = he->headend_links;
			assert(link->send_node == he &&
					link->recv_node == as->gateway);
			travel_link(link, args);
		}

		/* path: gateway->as_virt->outside */
		struct as_virt *virt_node = &as->virt_node;

		link = as->virt_link;
		assert(link->send_node == as && link->recv_node == virt_node);
		travel_link(link, args);

		/* path: as_virt->outside */

		/* from AS, look for receiver in univ or other AS */
		if (is_univ_node(receiver_id)) {
			/* path: as_virt->univ_virt->univ */
			link = &virt_node->univ_links[receiver_id];
			struct univ_node *univ = get_univ_node(receiver_id);
			assert(univ);
			struct univ_virt *univ_virt = &univ->virt_node;
			assert(link->send_node == virt_node &&
					univ_virt == link->recv_node);
			travel_link(link, args);
			/* path: univ_virt->univ */
			link = univ_virt->univ_link;
			assert(link->send_node == univ_virt &&
					link->recv_node == univ);
			travel_link(link, args);
			/* path complete */
		} else {
			/* path: as_virt->peer_as_virt->inside_as*/
			struct as *peer_as = NULL;
			unsigned int i;
			for (i = 0; i < _num_as; i++) {
				if (ID_IN_AS(receiver_id, &_as[i])) {
					peer_as = &_as[i];
					break;
				}
			}
			assert(peer_as && as != peer_as);
			if (as < peer_as) {
				link = &as->virt_node.gateway_links[i-1];
			} else {
				link = &as->virt_node.gateway_links[i];
			}
			assert(link->send_node == as &&
					peer_as == link->recv_node);
			as_link_traveler(peer_as,
					receiver_id,
					travel_link,
					args);
		}
	}
}

static void as_link_traveler(struct as *as,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args)
{
	assert(as);

	/* path: as_virt->as->inside_as */

	/* use the AS's virtual link */
	struct link *virt_link = as->virt_node.as_link;
	travel_link(virt_link, args);

	/* path: as->headend->headend?->cable_virt->cable_modem */

	/* now, to find which headend office it's in */

	struct headend_node *he = as->gateway;
	unsigned int i;
	for (i = 0; i < as->num_nodes; i++) {
		if (receiver_id >= he[i].low_client_id &&
				receiver_id <= he[i].hi_client_id) {
			he = &he[i];
			/* if it's the gateway, don't need to travel the link
			   since we're already at the gateway */
			if (he != as->gateway) {
				struct link *gateway_he_link =
				  &as->gateway->headend_links[i-1];
				assert(gateway_he_link->recv_node
						== he);
				travel_link(gateway_he_link, args);
			}
			break;
		}
	}
	assert(i < as->num_nodes);

	/* path: headend->cable_virt->cable_modem */
	headend_link_traveler(he,
			receiver_id,
			travel_link,
			args);
}

static __inline__ void headend_link_traveler(struct headend_node *he,
		unsigned int receiver_id,
		void (*travel_link)(struct link *link, void *args),
		void *args)
{
	assert(he && he->virt_link);
	assert(receiver_id >= he->low_client_id &&
			receiver_id <= he->hi_client_id);

	/* path: headend->cable_virt->cable_modem */
	travel_link(he->virt_link, args);

	/* path: cable_virt->cable_modem */
	unsigned int leaf_link_offset = receiver_id - he->low_client_id;
	struct link *link = &he->virt_node.cable_leaf_links[leaf_link_offset];
	struct cable_leaf_node *receiver = get_leaf_node(receiver_id);
	assert(link->recv_node == receiver);
	travel_link(link, args);

	/* path complete */
}

static __inline__ int is_univ_node(unsigned int client_id)
{
	return (client_id < _num_univ_nodes);
}

static __inline__ struct univ_node *get_univ_node(unsigned int client_id)
{
	assert(is_univ_node(client_id));
	return &_univ_node[client_id];
}

static __inline__ struct cable_leaf_node *get_leaf_node(unsigned int client_id)
{
	assert(!is_univ_node(client_id));
	assert(client_id < _num_univ_nodes + _num_cable_leaf_nodes);

	return &_cable_leaf_node[client_id - _num_univ_nodes];
}

static __inline__ unsigned int univ_get_client_id(struct univ_node *univ)
{
	assert(univ);

	return univ - _univ_node;
}

static __inline__ unsigned int leaf_get_client_id(struct cable_leaf_node *leaf)
{
	assert(leaf);

	return (leaf - _cable_leaf_node) - _num_univ_nodes;
}

static void add_conn_to_link(struct link *link, void *args)
{
	assert(link && args);
	struct conn *conn = (struct conn*)args;
	assert(link->connections.find(conn) == link->connections.end());

	link->connections.insert(conn);
}

static void remove_conn_from_link(struct link *link, void *args)
{
	assert(link && args);
	struct conn *conn = (struct conn*)args;

	assert(!link->connections.empty());
	conn_set::iterator it = link->connections.begin();
	
	assert(link->connections.find(conn) != link->connections.end());

	link->connections.erase(conn);
}

static __inline__ unsigned int link_match(struct link *link,
		void *send_node,
		void *recv_node)
{
	assert(link && send_node && recv_node);

	return (link->send_node == send_node && link->recv_node == recv_node);
}

void calc_data_transfers(unsigned int tick)
{
	DEBUG_MSG("Calculating data transfers on connections.\n");
	
	/* first initialize the link vector */
	assert(_link_vec.empty());
	for (unsigned int i = 0; i < _num_links; i++) {
		if (!_links[i].connections.empty()) {
			_links[i].bandwidth = _bandwidth[_links[i].type];
			assert(_links[i].conns_copy.empty());
			conn_set::iterator it = _links[i].connections.begin();
			for (; it != _links[i].connections.end(); it++) {
				struct conn *conn = (struct conn*)*it;
				assert(conn);
				_links[i].conns_copy.insert(conn);
				g_closeness[conn->closeness]++;
			}
			assert(_links[i].connections.size() ==
					_links[i].conns_copy.size());
			_link_vec.push_back(&_links[i]);
			/*
			DEBUG_MSG("link[%d]->bandwidth: %f\n",
					i,
					_links[i].bandwidth);
					*/
		}
	}

	/* enter bandwidth calculation loop */
	while (!_link_vec.empty()) {
		make_heap(_link_vec.begin(), _link_vec.end(), more_bandwidth());
		pop_heap(_link_vec.begin(), _link_vec.end());
		/*
		for (unsigned int i = 0; i < _link_vec.size(); i++) {
			DEBUG_MSG("link[%d], with %d"
					" connections and %f bytes/s/conn.\n",
					_link_vec[i] - _links,
					_link_vec[i]->conns_copy.size(),
					(float)_link_vec[i]->bandwidth /
					(float)_link_vec[i]->conns_copy.size());
		}
		*/
		struct link *link = _link_vec.back();
		assert(link);
		_link_vec.pop_back();
		if (!link->conns_copy.empty()) {
			float bw_per_conn = link->bandwidth /
				(float)link->conns_copy.size();
			/*
			DEBUG_MSG("Bottleneck link is link[%d], with %d"
					" connections and %f bytes/s/conn.\n",
					link - _links,
					link->conns_copy.size(),
					bw_per_conn);
			*/
			conn_set::iterator it;
			for (it = link->conns_copy.begin();
					it != link->conns_copy.end();
					it++) {
				struct conn *conn = *it;
				struct conn_bw conn_bw;
				assert(conn);
				conn_bw.conn = conn;
				conn_bw.bandwidth = bw_per_conn;
				link_traveler(conn->sender_id,
						conn->receiver_id,
						adjust_bandwidth,
						&conn_bw);
				unsigned int bytes = (unsigned int)
					(conn_bw.bandwidth *
					(float)(tick - conn->tick_update) /
					(float)TICKS_PER_SECOND);
				conn->tick_update = tick;
				struct bytes_transfer xfer;
				xfer.sender_id = conn->sender_id;
				xfer.receiver_id = conn->receiver_id;
				xfer.num_bytes = bytes;
				_xfer_vec.push_back(xfer);
			}
			link->conns_copy.clear();

			/* go through link vec and pop links with 0 conns */
			link_vector::iterator link_it;
			for (link_it = _link_vec.begin();
					link_it != _link_vec.end();) {
				link = (struct link*)*link_it;
				if (link->conns_copy.empty()) {
					link_it = _link_vec.erase(link_it);
				} else {
					link_it++;
				}
			}
		}
	}

	while (!_xfer_vec.empty()) {
		struct bytes_transfer xfer = _xfer_vec.back();
		_xfer_vec.pop_back();
		bt_peer *sender = &_peers[xfer.sender_id];
		bt_peer *receiver = &_peers[xfer.receiver_id];
		receiver->receive_bytes(sender, xfer.num_bytes);
	}

	update_bw_usage();

	struct event *ev = event_alloc();
	ev->fire_time = tick + TXEVENT_TICKS;
	ev->peer = NULL;
	ev->type = E_TRANSFERBLOCKS;
	event_insert(ev);
}

static void adjust_bandwidth(struct link *link, void *args)
{
	assert(link && args);
	struct conn_bw *conn_bw = (struct conn_bw*)args;

	link->conns_copy.erase(conn_bw->conn);
	if (!(conn_bw->bandwidth <= link->bandwidth + 100.0)) {
		fprintf(stderr, "%f %f\n", conn_bw->bandwidth, link->bandwidth);
	}
	assert(conn_bw->bandwidth <= link->bandwidth + 100.0);
	if (conn_bw->bandwidth > link->bandwidth) {
		link->bandwidth = 0;
	} else {
		link->bandwidth -= conn_bw->bandwidth;
	}
}

static void update_bw_usage()
{
	double old_percent = (double)_bw_log_events /
		(double)(_bw_log_events + 1);
	double new_percent = 1 / (double)(_bw_log_events + 1);
	
	for (unsigned int i = 0; i < _num_univ_nodes; i++) {
		bt_peer *peer = &_peers[i];
		if (!peer->is_active_peer) {
			continue;
		}
		struct univ_node *univ = &_univ_node[i];
		univ->virt_node.outgoing_bw_percent *= old_percent;

		double bw_used = _bandwidth[univ->virt_link->type] -
				univ->virt_link->bandwidth;
		double bw_avail = _bandwidth[univ->virt_link->type];

		univ->virt_node.outgoing_bw_percent +=
			new_percent * bw_used / bw_avail;


		univ->virt_node.incoming_bw_percent *= old_percent;
		bw_used = _bandwidth[univ->virt_node.univ_link->type] -
				univ->virt_node.univ_link->bandwidth;
		bw_avail = _bandwidth[univ->virt_node.univ_link->type];
		univ->virt_node.incoming_bw_percent +=
			new_percent * bw_used / bw_avail;
	}

	for (unsigned int i = 0; i < _num_cable_leaf_nodes; i++) {
		bt_peer *peer = &_peers[i+_num_univ_nodes];
		if (!peer->is_active_peer) {
			continue;
		}
		struct cable_leaf_node *leaf = &_cable_leaf_node[i];
		leaf->outgoing_bw_percent *= old_percent;

		double bw_used = _bandwidth[leaf->virt_link->type] -
				leaf->virt_link->bandwidth;
		double bw_avail = _bandwidth[leaf->virt_link->type];

		leaf->outgoing_bw_percent += new_percent *
			bw_used / bw_avail;

		leaf->incoming_bw_percent *= old_percent;
		struct cable_virt *virt_node = (struct cable_virt*)
			leaf->virt_link->recv_node;

		struct link *link = virt_node->cable_leaf_links;
		while (link->recv_node != leaf) {
			link++;
		}
		bw_used = _bandwidth[link->type] - link->bandwidth;
		bw_avail = _bandwidth[link->type];

		leaf->incoming_bw_percent += new_percent *
			bw_used / bw_avail;
	}

	for (unsigned int i = 0; i < _num_as; i++) {
		struct as_virt *virt_node = &_as[i].virt_node;
		virt_node->outgoing_bw_percent *= old_percent;
		double bw_used = _bandwidth[_as[i].virt_link->type] -
				_as[i].virt_link->bandwidth;
		double bw_avail = _bandwidth[_as[i].virt_link->type];
		virt_node->outgoing_bw_percent += new_percent *
			bw_used / bw_avail;
		virt_node->incoming_bw_percent *= old_percent;
		bw_used = _bandwidth[virt_node->as_link->type] -
				virt_node->as_link->bandwidth;
		bw_avail = _bandwidth[virt_node->as_link->type];
		virt_node->incoming_bw_percent += new_percent *
			bw_used / bw_avail;
	}

	for (unsigned int i = 0; i < _num_headend_nodes; i++) {
		struct headend_node *he = &_headend_node[i];
		struct cable_virt *virt_node = &he->virt_node;
		virt_node->outgoing_bw_percent *= old_percent;
		double bw_used = _bandwidth[virt_node->headend_link->type] -
				virt_node->headend_link->bandwidth;
		double bw_avail = _bandwidth[virt_node->headend_link->type];
		virt_node->outgoing_bw_percent += new_percent *
			bw_used / bw_avail;
		virt_node->incoming_bw_percent *= old_percent;
		bw_used = _bandwidth[he->virt_link->type] -
				he->virt_link->bandwidth;
		bw_avail = _bandwidth[he->virt_link->type];
		virt_node->incoming_bw_percent += new_percent *
			bw_used / bw_avail;
	}

	_bw_log_events++;
}

unsigned int get_num_external_peers()
{
	return _num_external_peers;
}

void log_bw_usage()
{
	log_msg("Bandwidth utilization statistics:\n\n");
	log_msg("\tUpload/Download bandwidth utilization for university nodes:\n");
	for (unsigned int i = 0; i < get_num_univ_nodes(); i++) {
		struct univ_virt *univ_virt = &_univ_node[i].virt_node;
		log_msg("\tuniv_node[%u]: (%4.2f%%,%4.2f%%)\n",
				i,
				univ_virt->outgoing_bw_percent * 100,
				univ_virt->incoming_bw_percent * 100);
	}
	log_msg("\n\tUpload/Download bandwidth utilization for cable modems:\n");
	for (unsigned int i = 0; i < _num_cable_leaf_nodes; i++) {
		struct cable_leaf_node *leaf = &_cable_leaf_node[i];
		log_msg("\tcable_node[%u]: (%4.2f%%,%4.2f%%)\n",
				i,
				leaf->outgoing_bw_percent * 100,
				leaf->incoming_bw_percent * 100);
	}
	log_msg("\n\tUpload/Download bandwidth utilization for headend shared links:\n");
	for (unsigned int i = 0; i < _num_headend_nodes; i++) {
		struct cable_virt *virt = &_headend_node[i].virt_node;
		log_msg("\theadend[%u]: (%4.2f%%,%4.2f%%)\n",
				i,
				virt->outgoing_bw_percent * 100,
				virt->incoming_bw_percent * 100);
	}
	log_msg("\n\tUpload/Download bandwidth utilization for AS external links:\n");
	for (unsigned int i = 0; i < _num_as; i++) {
		struct as_virt *virt = &_as[i].virt_node;
		log_msg("\tAS[%u]: (%4.2f%%,%4.2f%%)\n",
				i,
				virt->outgoing_bw_percent * 100,
				virt->incoming_bw_percent * 100);
	}
}
